import re

from .errors import ReferenceError
from .constants import book_regex

# TODO : document
# TODO : test
# TODO : create to_simple_string function?


def normalize_reference_string(reference_string):
    """Expand a reference string to full reference
    """
    return [
        ref.start.to_string() if ref.is_single_verse else ref.to_string()\
        for ref in parse_reference_string(reference_string)
    ]

def shorten_reference_string(reference_string):
    """Shorten a reference string

    Returns the minimum necessary string to unambigiously
    list references: E.g. `Genesis 1, Genesis 2` -> `Genesis 1, 2`
    """
    range_list = parse_reference_string(reference_string)
    if len(range_list) == 0:
        return ""

    string_list = [range_list[0].to_string()]
    if len(range_list) == 1:
        return string_list[0]

    for i in range(1,len(range_list)):
        string_list.append(range_list[i].to_string_short(range_list[i-1]))

    return ", ".join(string_list)


def parse_reference_string(reference_string):
    """Parse reference string to list of range objects
    """
    string_list = reference_string.strip().replace(";",",").split(",")
    res = [Range.from_string(string_list.pop(0))]
    for _string in string_list:
        res.append(Range.from_string(_string, res[-1]))
    return res

def _get_normalized_book_name(string):
    """Get the full book name from partial string
    """
    _string = string.strip()
    for regex in book_regex:
        if re.match(regex[0] + "[a-z\.]*", _string):
            return regex[1]
    raise ReferenceError("Cannot find book `{}`".format(_string))

class Reference:
    """A bible reference to either a chapter or verse
    """
    BCV_regex = re.compile("^([0-9]?[ _\.]?[A-Za-z]+)[\._ ]*([0-9]+)[\.: _]?([0-9]*)$")
    CV_regex = re.compile("^([0-9]+)[\.: ]([0-9]+)$")
    V_regex = re.compile("^([0-9]+)$")

    def __init__(self, book, chapter, verse = None):
        self.book = _get_normalized_book_name(book)
        self.chapter = int(chapter)
        try:
            self.verse = int(verse)
        except (TypeError, ValueError):
            # This reference is to the whole chapter
            self.verse = None

    def __repr__(self):
        return "<{}>".format(self.to_string())
    
    def to_string(self):
        if self.verse is None:
            return "{} {}".format(self.book,self.chapter)
        return "{} {}:{}".format(self.book,self.chapter,self.verse)
    
    def is_before(self,other):

        # Handle book comparision
        if self.book != other.book:
            raise ReferenceError("Cannot compare reference orders from different books")

        # Handle Chapter comparison
        if self.chapter < other.chapter:
            return True
        if other.chapter < self.chapter:
            return False
        
        # Handle verse comparisions
        if self.verse is None and other.verse is None:
            return True
        if other.verse is None or self.verse is None:
            raise ReferenceError("Cannot compare a chapter reference and a verse reference from same chapter")
        return self.verse < other.verse


    def equals(self,other):
        return self.book == other.book \
            and self.chapter == other.chapter \
            and self.verse == other.verse
    
    
    @classmethod
    def from_string(cls, string, previous = None):
        """Generate object from string (and previous obj if any)
        """
        _string = string.strip()
        match = cls.BCV_regex.match(_string)
        if not match is None:
            return cls(*match.groups())

        if previous is None:
            raise ReferenceError("Cannot infer book/chapter for reference")

        try:
            previous_book = previous.book
            previous_chapter = previous.chapter
            previous_verse = previous.verse
        except AttributeError:
            # previous is a Range object
            previous_book = previous.end.book
            previous_chapter = previous.end.chapter
            previous_verse = previous.end.verse

        match = cls.CV_regex.match(_string)
        if not match is None:
            # Infer book from previous
            return cls(previous_book, *match.groups())

        match = cls.V_regex.match(_string)
        if not match is None:
            if previous_verse is None:
                # previous is a chapter reference or range,
                #  so treat this like a chapter reference
                # TODO : single numbers after ranges spanning
                #  2+ chapters is ambigous: `John 3:10-4:12, 15`
                #  is that `John 15` or `John 4:15`? right now
                #  we treat that as the former.
                return cls(previous_book,*match.groups())
            # Infer both book and chapter from previous
            return cls(previous_book,previous_chapter,*match.groups())
        
        raise ReferenceError("Cannot parse string `{}` with previous `{}`".format(_string,previous.to_string()))

class Range:
    def __init__(self, start_reference, end_reference):
        try:
            if not (start_reference.is_before(end_reference) or start_reference.equals(end_reference)):
                raise ReferenceError("Cannot make range with end before beginning")
                # self.start = end_reference
                # self.end = start_reference
            else:
                self.start = start_reference
                self.end = end_reference
        except ReferenceError as e:
            raise ReferenceError("Cannot create range from provided references") from e
        # self.start = start_reference
        # self.end = end_reference

    def __repr__(self):
        return "<{}>".format(self.to_string())

    def to_string(self):
        return "{} - {}".format(self.start.to_string(), self.end.to_string())

    @property
    def book(self):
        return self.start.book

    @property
    def is_single_verse(self):
        return (not self.start.verse is None) and self.start.equals(self.end)

    @property
    def is_single_chapter(self):
        return (self.start.verse is None) and self.start.equals(self.end)

    def contains(self, ref):
        if self.start.book != ref.book:
            raise ReferenceError("Cannot compare references from different books")
        
        if self.start.verse is None and ref.chapter >= self.start.chapter:
            ge_start = True
        else:
            ge_start = not ref.is_before(self.start)
        
        if self.end.verse is None and ref.chapter <= self.end.chapter:
            le_end = True
        else:
            le_end = not self.end.is_before(ref)

        return ge_start and le_end

    def to_string(self):
        string = "{} {}".format(self.start.book, self.start.chapter)
        if not self.start.verse is None:
            string = string + ":{}".format(self.start.verse)
        if self.start.equals(self.end):
            return string

        string = string + "-"

        if self.end.chapter == self.start.chapter:
            return string + "{}".format(self.end.verse)

        if self.end.verse is None:
            return string + "{}".format(self.end.chapter)

        return string + "{}:{}".format(self.end.chapter, self.end.verse)

    def to_string_short(self, previous = None):
        """Generate the shortest reference string from this range
        """
        full = self.to_string()

        if previous is None or previous.start.book != self.start.book:
            return full

        # FIXME : assume Gen1:1-3:4,6 means also 3:6
        if (
            self.start.is_before(previous.end)\
            or previous.end.chapter != self.start.chapter\
            or self.start.chapter != self.end.chapter
        ):
            # Truncate only the book name
            return full.split(" ")[-1]

        return full.split(" ")[-1].split(":")[-1]


    @classmethod
    def from_string(cls, string, previous = None):

        strings = [_string.strip() for _string in string.split("-")]
        
        if len(strings) == 0 or len(strings) > 2:
            raise ReferenceError("Cannot create range from `{}`".format(string))

        ref = Reference.from_string(strings[0], previous)

        return cls(
            ref, 
            Reference.from_string(strings[len(strings) == 2], ref)
        )




import re
from bs4 import BeautifulSoup
import requests as r

from .errors import BadQueryError
from .reference import Reference
from .cleaner import clean_string

# TODO : clean up the strings to be asciii safe? i.e. \u2014 -> '--'
# TODO : error check the r.get 

class Verse(Reference):
    """Verse object for bible verses
    """
    def __init__(self, book, chapter, verse, version, text="", footnotes=None, crossrefs=None):
        super().__init__(book, chapter, verse)
        self.text = text
        self.version = version
    
        # need to intantiate empty list here to prevent all Verse instances
        #  sharing a reference to a single list object.
        if footnotes is None:
            self.footnotes = []
        else:
            self.footnotes = footnotes

        if crossrefs is None:
            self.crossrefs = []
        else:
            self.crossrefs = crossrefs

    def __repr__(self):
        return "<Verse: {}>".format(super().to_string())

    def to_string(self):
        return "{} `{}` ({})".format(super().to_string(),self.text,self.version)

    def to_dict(self):
        return dict(
            book = self.book,
            chapter = self.chapter,
            verse = self.verse,
            version = self.version,
            text = self.text,
            footnotes = self.footnotes,
            crossrefs = self.crossrefs
        )

    @classmethod
    def from_dict(cls,obj):
        return cls(**obj)

    def add_crossref(self, crossref):
        self.crossrefs.append(crossref)

    def add_footnote(self, footnote):
        self.footnotes.append(footnote)

    def equals(self, other):
        """Checks if provided Verse is equal to this one
        """
        if not isinstance(other,Verse):
            return False
        return self.version == other.version\
            and self.chapter == other.chapter\
            and self.book == other.book\
            and self.verse == other.verse

def _extract(obj, verse, sc):
    # If no 'attrs', then this is probably just a string
    #  so append it
    try:
        obj.attrs
    except:
        verse.text = verse.text + clean_string(obj)
        return

    # Ignore html tags that don't have a class
    try:
        obj.attrs['class']
    except:
        return
    # Recursive call on `words of jesus` to unpack
    if "woj" in obj.attrs['class']:
        for child in obj.children:
            _extract(child, verse, sc)

    # Get the tetragrammoton
    if "small-caps" in obj.attrs['class']:
        verse.text = verse.text + clean_string(obj.text).upper()
        return

    # Get all the footnotes
    if "footnote" in obj.attrs['class']:
        fn_id = obj.attrs['data-fn']
        fn = sc.find(id=fn_id[1:]).find(class_="footnote-text")
        html = ""
        for child in fn.children:
            try:
                child.attrs['class']
            except:
                html = html + clean_string(child)
            else:
                if "bibleref" in child.attrs['class']:
                    html = html + clean_string(child.text)
        # fn.attrs['class'] = "bible-footnote"
        verse.add_footnote(dict(
            str_index = len(verse.text),
            html = html#str(fn)
        ))
        return

    # Get all the crossreferences
    if "crossreference" in obj.attrs['class']:
        cr_id = obj.attrs['data-cr']
        cr = sc.find(id=cr_id[1:]).find(class_="crossref-link")
        verse.add_crossref(dict(
            str_index = len(verse.text),
            ref_list = [s.strip() for s in cr.attrs['data-bibleref'].split(",")]
        ))
        return


def scrap(ref_string_or_obj, version="ESV"):
    """Scrap a verse or chapter from the web
    Params
    ------
    ref_string_or_obj : str or Reference
        The string of a chapter/verse to query

    version (optional) : str
        The bible version to query from, default is "ESV"
        
    Returns
    -------
    Verse_or_list : Verse,list of Verses
        Either a list of, or just the single verse object in formatted json
    """
    
    # Ensure that we have a reference object
    if type(ref_string_or_obj) is str:
        ref = Reference.from_string(ref_string_or_obj)
    else:
        ref = ref_string_or_obj

    # Just in case `ref` is actually a `Verse` instance, infer version
    try:
        _version = ref.version
    except AttributeError:
        _version = version

    # Generate URL
    if ref.verse is None:
        URL = "https://www.biblegateway.com/passage/?search={}+{}&version={}".format(ref.book, ref.chapter, _version)
    else:
        URL = "https://www.biblegateway.com/passage/?search={}+{}:{}&version={}".format(ref.book, ref.chapter, ref.verse, _version)

    page = r.get(URL)
    
    sc = BeautifulSoup(page.content, "html.parser")

    if "No results found" in sc.text:
        raise BadQueryError("Cannot get {} ({})".format(ref.to_string(), _version))
    
    # Get all the p tags underneath the passage content
    #  which are all the verse, and no headings
    par = sc.find(class_="passage-content").find_all("p")

    # Infer all the available verses from the html class
    #  attribute of the `text` objects
    verse_ids = dict()
    for node in sc.find_all(class_="text"):
        for clsstr in node.attrs['class']:
            if re.match("[A-Za-z0-9]+[-]\d+[-]\d+",clsstr):
                vv = int(clsstr.split("-")[-1])
                verse_ids[vv] = clsstr

    # Generate a list of Verse instances
    verses = []
    for i,v in enumerate(verse_ids.keys()):
        
        # Get all the nodes matching the verse id
        elem_list = []
        for p in par:
            elem_list += p.find_all(class_=verse_ids[v])
        
        # Skip this round if no verses were found
        if len(elem_list) < 1:
            continue
        
        # Concatentate all the found verses (space delimited)
        #  since verses over carraige returns have mulitple
        #  html tags associated with them 
        children = []
        for node in elem_list:
            children += list(node.children)
            children += [" "]
        children.pop()

        # Prep verse object
        verse = Verse(ref.book, ref.chapter, v, _version)

        # Loop through all the found verse html tags, and extract the text
        for obj in children:
            # try:
            #     print(obj.text)
            # except:
            #     print(obj)
            _extract(obj, verse, sc)
        # append the completed verse
        if verse.text != "":
            verses.append(verse)

    if ref.verse is None:
        return verses
    return verses[0]




import click
import json

from pathlib import Path

from .query import query

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

def _format_line(obj, _format):
    if _format == 'refstr':
        return obj.to_string()
    if _format == 'str':
        return '`{}` ({})'.format(obj.text,obj.version)

def _formatter(obj_list, _format):
    # Handle json
    if _format == 'json':
        return json.dumps([v.to_dict() for v in obj_list],indent=4)

    # Handle everything else
    return "\n".join([_format_line(obj,_format) for obj in obj_list])

@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--version', '-v', default='ESV', help='Bible version to query, default is `ESV`')
@click.option('--format', '-f', '_format', default='refstr', type=click.Choice(['refstr','str','json']),
        help='Specify output format, default is `refstr`')
@click.option('--cache/--no-cache',  default=False, 
        help='Look up verses saved in a local cache first, and save new queries locally')
@click.argument('reference_string')
def scrap(version, reference_string, _format, cache):
    """Scrap bible verses

    REFERENCE_STRING a (comma delimited) list of references, e.g. `John3.16` or `1Peter3.1-5` or `Gen1,2`
    """
    
    verses = query( reference_string, version, cache )
    click.echo(_formatter(verses, _format))




import json

from .scraper import Verse

class JSONFile:
    """Abstraction layer for JSON file IO

    This class creates a (real or virtual) json file containing
    a single `{...}` objects, then allows the user to add things
    into, or read things out of the file, saving after each 
    write operation

    Parameters
    ----------
    fp : string or None
        A file path to where the JSON file exsits, or where it should
        be immediately created. If set to `None` the class will create
        and empty "virtual" file, which will be destroyed with the 
        class instance. Default is `None`.
    """
    def __init__(self, fp = None):
        self.fp = fp
        self.json = dict()
        self.pull()

    @property
    def file_path(self):
        if self.fp is None:
            return "Virtual"
        return self.fp

    def __repr__(self):
        return "<JSONFile @ {}>".format(self.file_path)

    def pull(self):
        """Load contents of the JSON file into memory

        Note, this will destroy any unsaved changes currently
        in the class instance memory. (Use `.push` to save)
        """
        # Virtual files don't get saved to the hard disk
        if self.fp == None:
            return self

        # Attempt to open file
        try:
            f = open(self.fp)
        except FileNotFoundError:
            # The files doesn't exist yet, so nuke any local
            #  json configuration
            self.json = dict()
            return self

        # Load json, might throw an error if your file is
        #  corrupted
        self.json = json.load(f)
        f.close()

        return self

    def push(self):
        """Save local json object to the hard disk

        Note, this creates a file at `self.fp` if one
        does not already exist there.
        """
        # Virtual files don't get saved anywhere
        if self.fp == None:
            return self

        # Save, might throw an error if you did something
        #  weird to `self.json`
        with open(self.fp, "w+") as f:
            json.dump(self.json, f)

        return self

class BibleCacher:
    """Abstraction layer for bible verse/range IO to files

    This class allows bible verse (previously queried) to
    be cached locally in a file, so that repeated queries
    can just look up the result.

    The file structure is as follows:
    ```json
    {
        "queries" : {
            "Normalized range string" : "List of verse strings in range",
            "John 3:16-17 (ESV)" : [
                "John 3:16 (ESV)",
                "John 3:17 (ESV)"
            ],
            ...
        },
        "items" : {
            "Normalized verse string" : "serialized `bsw.scraper.Verse` object",
            "John 3:16 (ESV)" : { VerseObject },
            ...
        }
    }
    ```

    Parameters
    ----------
    fp : string or None
        A file path to where to cache bible verse, if the
        path doesn't exist, this will attempt to create it.
        If `None` then a virtual file will be used, and 
        also destroyed with this class instance. Default
        is `None`.
    """
    def __init__(self, fp = None):
        self.file = JSONFile(fp)
        
        # Setup structure
        changed = False
        try:
            self.json['queries']
        except KeyError:
            self.json['queries'] = dict()
            changed = True

        try:
            self.json['items']
        except KeyError:
            self.json['items'] = dict()
            changed = True

        if changed:
            self.file.push()

    @property
    def json(self):
        return self.file.json

    @property
    def file_path(self):
        return self.file.file_path

    def __repr__(self):
        return "<BibleCacher @ {}>".format(self.file_path)

    def generate_query_key(self, range_or_range_string, version):
        """Generate key for a `query` (list of normalized verse strings)
        """
        if type(range_or_range_string) is str:
            query_key = range_or_range_string + " ({})".format(version.upper())
        else:
            # Treat `range_string` as a range object and stringify it
            query_key = range_or_range_string.to_string() + " ({})".format(version.upper())

        return query_key

    def generate_item_key(self, verse):
        """Generate the key for an `item` (serialized verse object)
        """
        return "{} {}:{} ({})".format(verse.book, verse.chapter, verse.verse, verse.version)

    def look_up(self, range_or_range_string, version="ESV"):
        """Check the cache file for the result of a bible range
        """

        query_key = self.generate_query_key( range_or_range_string, version )
        try:
            item_key_list = self.json['queries'][query_key]
        except KeyError:
            return None

        return [ Verse.from_dict(self.json['items'][item_key]) for item_key in item_key_list ]

    def cache(self, range_or_range_string, verse_list):
        """Save the results of a query to the cache
        """

        query_key = self.generate_query_key( range_or_range_string, verse_list[0].version )
        self.json['queries'][query_key] = []
        for verse in verse_list:
            item_key = self.generate_item_key(verse)
            self.json['queries'][query_key].append(item_key)
            self.json['items'][item_key] = verse.to_dict()

        self.file.push().pull()

        return self

